local preset = rawget(_G, 'preset')
preset = preset ~= nil and preset ~= '' and preset or 'ars'

local defaults = {
    -- read-only modes:
    playable_mode = false,
    playback_mode = false,

    -- controlling in-game mechanics:
    starting_world = 1, -- set to 0 to randomize the world every epoch.
    starting_level = 1, -- set to 0 to randomize the level every epoch.
    start_big = false,
    starting_lives = 0,
    min_time = 100,
    max_time = 300,
    timer_loser = 1/2, -- decrement the timer when mario is slacking off.

    -- rewards:
    decrement_reward = false, -- bad idea, encourages mario to run into goombas.
    score_multiplier = 1, -- how much the ingame score influences our rewards.

    -- network initialization:
    init_zeros = false, -- instead of random normal noise.

    -- network inputs (connections):
    time_inputs = true, -- insert binary inputs of a frame counter.

    -- network layers:
    hidden = false, -- use a hidden layer with ReLU/GELU activation.
    hidden_size = 128,
    layernorm = false, -- use a LayerNorm layer after said activation.
    reduce_tiles = false,
    bias_out = true,

    -- network evaluation (sampling joypad):
    frameskip = 4,
    -- true greedy epsilon has both deterministic and det_epsilon set.
    deterministic = false, -- use argmax on outputs instead of random sampling.
    det_epsilon = false, -- take random actions with probability eps.

    -- evolution strategy and non-rate hyperparemeters:
    es = 'ars',
    ars_lips = false, -- for ARS.
    epoch_top_trials = 9999, -- for ARS.

    -- sampling:
    deviation = 1.0,
    unperturbed_trial = true, -- perform an extra trial without any noise.
                              -- this is good for logging, so i'd recommend it.
    epoch_trials = 50,
    graycode = false, -- for ARS.
    negate_trials = true, -- try pairs of normal and negated noise directions.
    -- AKA antithetic sampling. note that this doubles the number of trials.
    min_refresh = 0.1, -- for SNES.

    -- epoch-related rates:
    base_rate = 1.0, -- param_rate, sigma_rate, and covar_rate will base on this
                     -- if you don't specify them individually.
    param_decay = 0.0,
    sigma_decay = 0.0, -- for SNES, xNES.
    momentum = 0.0, -- for ARS.
}

local presets = require("presets")

for _, cfg in pairs(presets) do
    local parent = defaults
    if cfg.parent ~= nil then
        parent = presets[cfg.parent]
        assert(parent, "no such parent preset: "..tostring(cfg.parent))
    end
    setmetatable(cfg, {__index=parent})
end

local cfg = presets[preset]
assert(cfg, "invalid preset: "..tostring(preset))

-- TODO: so, uhh..
-- what happens when playback_mode is true but unperturbed_trial is false?

if cfg.es == 'ars' then
    if cfg.param_rate == nil then cfg.param_rate = cfg.base_rate end
else
    if cfg.param_rate == nil then cfg.param_rate = 1.0 end
end
if cfg.sigma_rate == nil then cfg.sigma_rate = cfg.base_rate end
if cfg.covar_rate == nil then cfg.covar_rate = cfg.sigma_rate end

local parent = getmetatable(cfg).__index

setmetatable(cfg, {
    __index = function(t, n)
        if parent ~= nil and parent[n] ~= nil then return parent[n] end
        if n == 'name' then return nil end
        if n == 'log_fn' then return nil end
        if n == 'params_fn' then return nil end
        if n == 'stats_fn' then return nil end
        error("cannot use undeclared config '" .. tostring(n) .. "'", 2)
    end
})

cfg.epoch_top_trials = math.min(cfg.epoch_trials, cfg.epoch_top_trials)

cfg.eps_start = 1.0 * cfg.frameskip / 64
cfg.eps_stop  = 0.1 * cfg.eps_start
cfg.eps_frames = 1000000
cfg.enable_overlay = cfg.playable_mode
cfg.enable_network = not cfg.playable_mode

assert(not cfg.ars_lips or cfg.unperturbed_trial,
       "cfg.unperturbed_trial must be true to use cfg.ars_lips")
assert(not cfg.ars_lips or cfg.negate_trials,
       "cfg.negate_trials must be true to use cfg.ars_lips")
assert(not (cfg.es == 'snes' and cfg.negate_trials),
       "cfg.negate_trials is not yet compatible with SNES")

return cfg
